//********************************************************************
//*** ARDUINO 2 (PD2) 		 
//*** ARDUINO 12 (PB4) 		    
//*** ARDUINO 13 (PB5) 		    
//***
//*** CONFIG STRING FROM LabView:
//*** *IDN?\r 									--  
//*** *RESET\r									--    
//*** *STATUS?\r								--  
//*** *RUN:ONCE WIDTH=xxxxx\r					--     xxxxx ms
//*** *RUN:CYCLE PERIOD=xxxx WIDTH=yyyy COUNT=zzz\r		--      xxx ms,   yyy,   zzz (0 = )
//*** *RUN:SYNC RAISE=0|1 MINPERIOD=XXX MAXPERIOD=YYY DELAY=ZZZ WIDTH=AAA\r	-- 
//***

#define F_CPU 16000000UL
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//******************************************************************************
//******************************************************************************
//     *************************************
//******************************************************************************
//******************************************************************************

static int t1Count = 65535 - 15625 / 2;	//   1   LED 13
volatile unsigned char ledFlag;			//  -  LED  

static int t0Count = 65535 - 250;		//   0  1 
volatile unsigned long mSeconds;		//    .
volatile unsigned char Seconds;			//   
volatile unsigned char Minutes;			//   
volatile unsigned char Hours;			//   
volatile unsigned int  Days;			//   

enum states { ERROR = 0, IDLE = 1, RUNONCE, RUNCYCLE, RUNSYNC };
static int generatorState;				//   

volatile unsigned int configSignal;		// 0 = 0->1; 1 = 1->0
volatile unsigned int configMinPeriod;	// minimal period between signal
volatile unsigned int configMaxPeriod;	// maximal period between signal
volatile unsigned int configDelay;		// 100th mks
volatile unsigned int configWidth;		// 100th mks

volatile unsigned long prevTime;
volatile unsigned char enablePulse;
volatile unsigned char generatingPulse;



#define BUFFER_UART_SIZE 128
volatile uint8_t Nbyte = 0;
volatile char buffUART[BUFFER_UART_SIZE]; 
volatile int lvStringReady;


//******************************************************************
//   ,    
//******************************************************************
void baseConfig()
{
  configSignal = 0;				//   
  configMinPeriod = 0;			//    0
  configMaxPeriod = 4000;		//    4 
  configDelay = 10;				//   1000 
  configWidth = 10;				//   100 
}

//*******************
//   UART
//*******************
void ClearBuffUart(void)
{
    for(int i=0;i<BUFFER_UART_SIZE;i++)
	{
        buffUART[i]=0;
   	}   
    Nbyte = 0;
}


//******************************************
//      
//      
//******************************************
ISR( TIMER0_OVF_vect )
{
  cli();			//     
  mSeconds++;		//   
  TCNT0 = t0Count;	//   
  if (mSeconds % 1000 == 0) Seconds++;
  sei();			// 
  //          
  if (Seconds == 60) { Seconds = 0; Minutes++; }
  if (Minutes == 60) { Minutes = 0; Hours++; }
  if (Hours == 24) { Hours = 0; Days++; }
}

//********************************************
//      
//********************************************
ISR( TIMER1_OVF_vect )
{
  cli();//     
  if (ledFlag == 0) { PORTB |= 0b00100000; ledFlag = 1; } else { PORTB &= 0b11011111; ledFlag = 0; }
  TCNT1 = t1Count; 			//   TCNT1
  sei();// 
}


//**********************************************************************************************
//         (~1100  )
//**********************************************************************************************
unsigned long mSecDiff(unsigned long prevValue)
{
  unsigned long Value = mSeconds;

  if (prevValue > Value)
  {
    return (0xFFFFFFFF - prevValue + Value + 1);
  }
  else 
    return (Value - prevValue);
}

//********************************************************
//       2 (PD2)
//********************************************************
ISR(INT0_vect) //  INT0
{
  unsigned long diff;
 
  cli();//     
  
  // ,       . 
  //       
  //                
  if (prevTime != 0)
    diff = mSecDiff(prevTime);
  else
    diff = configMinPeriod;
  prevTime = mSeconds;
  if ((diff >= configMinPeriod) && (diff <= configMaxPeriod))
  {
  	//     ,     
    if (generatingPulse == 0) enablePulse = 1;
  }
 
  sei();// 
}


//*********************************************************************************************
//       PD2 (INT0)   
//*********************************************************************************************
void InitInt0()
{
  EIMSK |= 1<<INT0; 	//  INT0

  if (!configSignal)
    EICRA |= (1<<ISC01)|(1<<ISC00);	// 0: 0->1
  else
  {
    EICRA |= (1<<ISC01);	// 1: 1->0
	EICRA &= ~(1<<ISC00);
  }
  prevTime = 0;
  enablePulse = 0;
  generatingPulse = 0;
}

void DeInitInt0()
{
  EIMSK = 0; 	//  INT0
  prevTime = 0;
  enablePulse = 0;
  generatingPulse = 0;
}


//**************************************************************
//       RS232
//**************************************************************
void usartInit()
{
    UCSR0B = (1<<RXCIE0)|(1<<TXCIE0)|(1<<RXEN0)|(1<<TXEN0);
    UCSR0C = (1<<UCSZ01)|(1<<UCSZ00); //  8 	
	if (UCSR0A & (1<<U2X0)) //  -    
	{ 
	    UBRR0L = F_CPU / 4800 / 16 - 1;
    	UBRR0H = (F_CPU / 4800 / 16 - 1) >> 8;
	}
	else
	{ 
	    UBRR0L = F_CPU / 9600 / 16 - 1;
    	UBRR0H = (F_CPU / 9600 / 16 - 1) >> 8;
	}
	ClearBuffUart();
}

 
//******************************************
//      UART
//******************************************
void usartSendChar( char data )
{
    // Wait for empty transmit buffer 
    while ( !( UCSR0A & (1<<UDRE0)) );
    // Put data into buffer, sends the data 
    UDR0 = data;
}
 
//******************************************
//      UART
//******************************************
void usartSendString(char * data)
{
    unsigned char d;
    while(*data)
	{
        d = *data;
		data++;
        usartSendChar(d);
    }
}
 

//**********************
//   
//**********************
char usartReceive( void )
{
    // Wait for data to be received 
    while ( !(UCSR0A & (1<<RXC0)) );
    // Get and return received data from buffer 
    return UDR0;
}
 
//**************************************
//     USART
//**************************************
ISR(USART_TX_vect)
{
}

ISR(USART_RX_vect)
{
    unsigned char b;

    b = UDR0;

    //   ,      '*' 
	if ((Nbyte == 0) && (b != '*'))
	{
	  return;
	}

	//       ,     
	if (lvStringReady != 0) 
	{ 
        usartSendString("ERR1: unsolicided command");
		return; 
	}

	//    
    buffUART[Nbyte++] = b;

	//   
    if (Nbyte >= BUFFER_UART_SIZE)
	{
        usartSendString("ERR0: too long command");
        ClearBuffUart();
		lvStringReady = 0;
		return;
    } 
	
	//   ,      -.
	if (b == '\r')
	{
		buffUART[Nbyte] = 0;
		lvStringReady = 1;
	}
	
}

//**************************************
//    
//**************************************
void setGeneratorState(int state)
{
	generatorState = state;
    switch(generatorState)
	{
		case ERROR:	  t1Count = 65535 - 15625 / 8;	// ultra high speed blinking when error
		break;
		case IDLE:	  { t1Count = 65535 - 15625 / 1; baseConfig(); DeInitInt0(); }	// low speed blinking when running IDLE
		break;
		case RUNSYNC: { t1Count = 65535 - 15625 / 4; InitInt0(); }	// higher speed blinking when running SYNC
		break;
		default:	  t1Count = 65535 - 15625 / 2;	// high speed blinking when running ONCE or CYCLE
		break;
	}

}



//******************
//  
//******************
void processCommand()
{
	int i;
	char buf[128];

	if (strcmp((char *) buffUART, "*IDN?\r") == 0) 
	{
        usartSendString("ARDUINO 328P BASED GENERATOR\n");
		ClearBuffUart();
		lvStringReady = 0;
		return;
	}
	if (strcmp((char *) buffUART, "*RESET\r") == 0) 
	{
        usartSendString("RESET OK\n");
		ClearBuffUart();
		baseConfig();
		lvStringReady = 0;
		return;
	}
	if (strcmp((char *) buffUART, "*STATUS?\r") == 0) 
	{
	    switch(generatorState)
		{
			case ERROR:	  		usartSendString("ERROR\n");
			break;
			case IDLE:	  		usartSendString("IDLE\n");
			break;
			case RUNONCE:	  	usartSendString("RUN ONCE\n");
			break;
			case RUNCYCLE:	  	usartSendString("RUN CYCLE\n");
			break;
			case RUNSYNC:	  	usartSendString("RUN SYNCRONIZED\n");
			break;
			default:	  		usartSendString("UNKNOWN\n"); 
			break;	
		}
		ClearBuffUart();
		lvStringReady = 0;
		return;
	}

	if (strstr((char *) buffUART, "*RUN:ONCE") == buffUART)
	{
		if ((i = sscanf((char *) buffUART, "*RUN:ONCE WIDTH=%d", &configWidth)) == 1)
		{
			if ((configWidth > 0) && (configWidth <= 10000))
			{
				ClearBuffUart();
				lvStringReady = 0;
   	    		usartSendString("OK");
				setGeneratorState(RUNONCE);
			}
			else
			{
				sprintf(buf, "ERR: (RUN:ONCE) bad pulse width %d", configWidth);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
			}
		}
		else
		{
			baseConfig();
    	    usartSendString("ERR: Error in RUN:ONCE configuration string [*RUN:ONCE WIDTH=%d]");
			ClearBuffUart();
			lvStringReady = 0;
		}
		return;
	}

	if (strstr((char *) buffUART, "*RUN:CYCLE") == buffUART)
	{
		if ((i = sscanf((char *) buffUART, "*RUN:CYCLE WIDTH=%d PERIOD=%d COUNT=%d", &configWidth, &configMaxPeriod, &configMinPeriod)) == 3)
		{
			
			if ((configWidth <= 0) || (configWidth > 30000))
			{
				sprintf(buf, "ERR: (RUN:CYCLE) bad pulse width %d", configWidth);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configMaxPeriod <= configWidth) || (configMaxPeriod > 32000))
			{
				sprintf(buf, "ERR: (RUN:CYCLE) bad pulse period %d", configMaxPeriod);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configMinPeriod < 0 ) || (configMinPeriod > 30000))
			{
				sprintf(buf, "ERR: (RUN:CYCLE) bad pulse count %d", configMaxPeriod);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			ClearBuffUart();
			lvStringReady = 0;
   	    	usartSendString("OK");
			configDelay = configMaxPeriod - configWidth;
			setGeneratorState(RUNCYCLE);
		}
		else
		{
			baseConfig();
    	    usartSendString("ERR: Error in RUN:CYCLE configuration string [*RUN:CYCLE WIDTH=%d PERIOD=%d COUNT=%d]");
			ClearBuffUart();
			lvStringReady = 0;
		}
		return;
	}

	if (strstr((char *) buffUART, "*RUN:SYNC") == buffUART)
	{
		if ((i = sscanf((char *) buffUART, "*RUN:SYNC WIDTH=%d DELAY=%d RAISE=%d MIN=%d MAX=%d", &configWidth, &configDelay, &configSignal, &configMinPeriod, &configMaxPeriod)) == 5)
		{
			if ((configWidth <= 0) || (configWidth > 32000))
			{
				sprintf(buf, "ERR: (RUN:SYNC) bad pulse width %d", configWidth);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configDelay < 0) || (configDelay > 32000))
			{
				sprintf(buf, "ERR: (RUN:SYNC) bad pulse delay %d", configDelay);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configSignal != 0) && (configSignal != 1))
			{
				sprintf(buf, "ERR: (RUN:SYNC) bad raise/fall type %d", configSignal);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configMinPeriod < 1) || (configMinPeriod > 32000))
			{
				sprintf(buf, "ERR: (RUN:SYNC) bad min sync-period %d", configMinPeriod);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			if ((configMaxPeriod <= configMinPeriod) || (configMaxPeriod > 32000))
			{
				sprintf(buf, "ERR: (RUN:SYNC) bad max sync-period %d", configMaxPeriod);
	    	    usartSendString(buf);
				ClearBuffUart();
				lvStringReady = 0;
				baseConfig();
				return;
			}
			ClearBuffUart();
			lvStringReady = 0;
   	    	usartSendString("OK");
			enablePulse = 0;
			setGeneratorState(RUNSYNC);
		}
		else
		{
			baseConfig();
    	    usartSendString("ERR: Error in RUN:SYNC configuration string [*RUN:SYNC WIDTH=%d DELAY=%d RAISE=%d MIN=%d MAX=%d]");
			ClearBuffUart();
			lvStringReady = 0;
		}

		return;
	}

    sprintf(buf, "UNKNOWN COMMAND (%s)\n", buffUART);
	usartSendString(buf);
	
	ClearBuffUart();
	lvStringReady = 0;


}
 


//*******************************************************
//     
//*******************************************************
void Init()
{
  DDRB =  0xFF;		// PORT B  
  DDRD = 0x00;		// PORT D  
  PORTD = 0xFF;		//   D

  cli();

  ledFlag = 0;
  mSeconds = 0;
  Seconds = 0;
  Minutes = 0;
  Hours = 0;
  Days = 0;
  lvStringReady = 0;
  setGeneratorState(IDLE);


  //     T1
  TCCR1B |= (1<<CS12)|(0<<CS11)|(1<<CS10); //    1024
  TIMSK1 |= (1<<TOIE1); //     
  TCNT1 = t1Count;        //    TCNT1      

  //        T0
  TCCR0B |= (0<<CS02)|(1<<CS01)|(1<<CS00); //    64
  TIMSK0 |= (1<<TOIE0); //     
  TCNT0 = t0Count;        //    TCNT0    250  (1 )

  baseConfig();
  InitInt0();
    
  
  usartInit();

  sei();  
  return;
}

//************************************
// ,   
//************************************
int main(void)
{
  int i;

  Init();

  while(1)
  {
    //     UART  - ,     
	if (lvStringReady == 1) 
	{ 
		processCommand(); 
	}

	//   
	// RUN:ONCE
	//          configWidth
	if (generatorState == RUNONCE)
	{
	  generatingPulse = 1;
      PORTB |= 0b00010000;
	  for (i = 0; i < configWidth; i++) _delay_ms(1);
      PORTB &= 0b11101111;
	  generatingPulse = 0;
	  setGeneratorState(IDLE);
	  continue;
	}

	//   
	// RUN:CYCLE
	//      configMinPeriod    configWidth    configMaxPeriod
	if (generatorState == RUNCYCLE)
	{
	  generatingPulse = 1;
      PORTB |= 0b00010000;
	  for (i = 0; i < configWidth; i++) _delay_ms(1);
      PORTB &= 0b11101111;
	  for (i = 0; i < configDelay; i++) _delay_ms(1);
	  generatingPulse = 0;
	  //            
	  if (configMinPeriod == 0) continue;

	  //     ,         0     .
	  configMinPeriod--;
	  if (configMinPeriod == 0) { setGeneratorState(IDLE); }
	  continue;
	}

	//   
	// RUN:SYNC
	//       RESET   
    if ((generatorState == RUNSYNC) && (enablePulse == 1))
	{
	  generatingPulse = 1;
	  for (i = 0; i < configDelay; i++) _delay_us(100);
      PORTB |= 0b00010000;
	  for (i = 0; i < configWidth; i++) _delay_us(100);
      PORTB &= 0b11101111;
	  enablePulse = 0;
	  generatingPulse = 0;
	  continue;
	}

  }
  return 0;
}

